% Smith_Barnwell
%
% Este programa ejemplifica cmo disear el filtro FIR de reconstruccin perfecta,
% cmo procesar las seales con el banco qmf de anlisis,
% cmo cuantificar las seales con un nmero dado de bits
% y cmo reconstruir las seales con el banco de filtros de sntesis.
% Se verifica la reconstruccin perfecta
%
% Marco A. Alzate, DSP2
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Primero disea el filtro prototipo H(z) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[N, Fo, Ao, W] = firpmord([0.44 0.56],[1 0],[0.002 0.002]);   % Parametros de diseo del filtro
L=ceil((N-2)/4); N=4*L+2;									  % Ajusta a un orden adecuado
h = firpm(N,Fo,Ao,W);                                         % Filtro Prototipo H(z)
h(2:2:N)=zeros(size(h(2:2:N))); h(2*L+2)=1/2;                 % Asegura la media banda
h = (h + fliplr(h))/2;                                        % Asegura fase cero
H = abs(fft(h,16384)); d = 1.01*max(H(ceil(0.57*8192):8191)); % Encuentra el verdadero rizado
N = 2*L+1;                                                    % Orden del filtro h0
h(N+1)=0.5 + d;                                               % Asegura respuesta no negativa
w = pi*(0:1023)/1024; H=zeros(size(w));                       % Rango de frecuencias por graficar
for i=1:1024                                                  % Evalua DTFT con n= -N,1-N,...,N-1,N
   e=exp(1j*w(i)*(-N:N));                                     % para verificar que es un filtro de fase cero
   H(i)=h*e';                                                 % con respuesta no negativa
end
subplot(221); plot(w,real(H));  title('Re[H(w)]'); axis tight % Parte real (igual a la magnitud)
subplot(222); plot(w,imag(H));  title('Im[H(w)]'); axis tight % Parte imaginaria (igual a cero)
subplot(223); plot(w,abs(H));   title('|H(w)|'); axis tight   % Magnitud (igual a la parte real)
subplot(224); plot(w,angle(H)); title('phi[H(w)]'); axis tight % Fase (igual a cero)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Despues disea el filtro pasabajo de anlisis %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
zh=roots(h);                                                  % Raices de h(1)*z^2N + ... + h(2N+1)
figure; subplot(221); zplane(zh); title('Raices de H(z)')
z0=zh(find(abs(zh)<1));                                       % Ceros de fase mnima (dentro del circulo unitario)
h0=1;                                                         % Expandimos el producto como polinomio de z^-1.
for i=1:N, h0=conv(h0,[-z0(i)  1]); end                       % Prod(1-zk*z^-1)
h0=real(h0);                                                  % Algunos componentes imaginarios del orden 10^-16
h0=sqrt(2)*h0/sum(h0);                                        % Normaliza para ganancia sqrt(2) en dc
[H0,w]=freqz(h0,1,1024);                                      % Calcula la respuesta en frecuencia
subplot(222); plot(w,abs(H0)); axis([0 pi 0 1.5]); title('|H_0(w)|')             % y grafica su magnitud y fase
subplot(223); plot(w,angle(H0)); axis([0 pi -4 4]); title('phi[H_0(w)]')         % No se ve mal (casi lineal)
subplot(224); plot(w,20*log10(abs(H0))); axis([0 pi -50 10]); title('|H_0(w)|'); % En dB

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Despues sintetiza los otros filtros %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
f0=fliplr(h0);                                                % Pasabajos de sntesis
h1=((-1).^(0:N)).*f0;                                         % Pasaaltos de anlisis
f1=((-1).^(1:(N+1))).*h0;                                     % Pasaaltos de sntesis
save filtros_QMF_PR.mat h0 h1 f0 f1
[H1,w]=freqz(h1,1,1024);                                      % Respuestas en frecuencia
[F0,w]=freqz(f0,1,1024);
[F1,w]=freqz(f1,1,1024);
figure
subplot(211); plot(w,[abs(H0) abs(H1)]); axis([0 pi 0 1.5]); title('|H_i(w)|')
subplot(212); plot(w,[abs(F0) abs(F1)]); axis([0 pi 0 1.5]); title('|F_i(w)|')
T = (H0.*F0 + H1.*F1)/2;                                      % Funcin de transferencia
figure
subplot(211); plot(w,abs(T)); title('Funcin de transferencia')
axis([0 pi 0 1.1])

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Analiza y sintetiza una seal %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
x = sin(2*pi*(0:99)'/100) + 0.2*(rand(100,1)-0.5);              % Seal a procesar
x0 = upfirdn(x,h0,1,2);                                         % Salida del pasabajos de anlisis
x1 = upfirdn(x,h1,1,2);                                         % Salida del pasa-altos de anlisis
y0 = upfirdn(x0,f0,2,1);                                        % Salida del pasabajos de sntesis
y1 = upfirdn(x1,f1,2,1);                                        % Salida del pasaaltos de sntesis
y = y0+y1;                                                      % Salida del decodificador subbanda
subplot(212); plot(x); hold on; plot(y(N+(1:100)),'r--')        % Reconstruccin Perfecta!!
title('Reconstruccin Perfecta!!')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Por ltimo, comprime una seal de voz %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[x,fs]=audioread('Vocal_A.wav');                              % Seal a analizar: fonema /a/ (fs=8000, nb=8)
x0=upfirdn(x,h0,1,2);                                         % x0 va de -1.067V a 1.244V (2.311V)
x1=upfirdn(x,h1,1,2);                                         % x1 va de -0.122V a 0.174V (0.296V)
Delta = max(max(x0),-min(x0))*2^-7;                           % Paso de cuantizacin con 8 bits
NB1=log2(floor(2*max(max(x1),-min(x1))/Delta));               % Los detalles requieren 5.13 bits Lo redondearemos a 5.
                                                              % Esto implica prdida de informacin
x0q = round(x0/Delta); x0q=max(-128,min(127,x0q));            % Cuantizacin de las seales a a transmitir
x1q = round(x1/Delta); x1q=max( -16,min( 15,x1q));            % (Son nmeros enteros en [-128 +127] y [-16 15]
SNR0=10*log10((x0'*x0)/((x0-Delta*x0q)'*(x0-Delta*x0q)));     % El problema es que la cuantizacin nos genera ruido
SNR1=10*log10((x1'*x1)/((x1-Delta*x1q)'*(x1-Delta*x1q)));     % (SNR de 44 dB en x0 y de 23 dB en x1)
n=0:204; figure                                               % Indice de tiempo
subplot(221);stairs(n,x0q); hold on; plot(n,x0/Delta,'r--'); title('Cuantizacin de x_0[n]');axis tight
subplot(223);stairs(n,x1q); hold on; plot(n,x1/Delta,'r--'); title('Cuantizacin de x_1[n]');axis tight
subplot(222);plot(n,(x0-Delta*x0q).^2); title('Error de cuantizacin en x_0[n]');axis tight
subplot(224);plot(n,(x1-Delta*x1q).^2); title('Error de cuantizacin en x_1[n]');axis tight
y0=upfirdn(Delta*x0q,f0,2,1);                                 % De todas formas nos ahorramos 1.5fs = 12000 bps
y1=upfirdn(Delta*x1q,f1,2,1);                                 % (de 64kbps a 52 kbps con "la misma calidad")
z=y0(N+(1:384))+y1(N+(1:384));                                % Seal sintetizada
e=(x-z).^2; MSE=mean(e);                                      % N = 1.25E-5 W, S = 39.9 W ==> SNR = 65 dB
n=0:383; figure                                               % Esta alta SNR implica reconstruccin perfecta,
subplot(211); plot(n,x,'b-',n,z,'r:'); title('Entrada/Salida')% como se puede apreciar en esta figura.
axis tight                                                    % (Sin cuantizacin obtendramos SNR = 77 dB, donde
subplot(212); plot(n,e); hold on; plot([0 383],MSE*[1 1],'r:')% los errores son slo por aproximaciones numricas. De
title('Energia del error y error cuadrado medio'); axis tight % hecho, los tres picos del error son por clipping en x0)
